/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*    
      bubble.c - bubble functions and routines


*/

#include <math.h>
#include <stdio.h>
#include <glib.h>

//#include <glib.h>

#include "bubble.h"

/* KUNDU says air-water surface tension @ 20C */

const double surface_tension = 0.0728;   // N/m

/* FIX FIX FIX */

const double hinze_critical_radius_multiplyer=1.0 ;  // = ??????

const double alphas =   1.259921049894873;   // 2^(1/3); 

/* Global bubble variables in defs.h  */

/* scalar variables bubble radius stuff  */

int nr;         // number of bubble radius
double dr;      // see defs.h for info
double idr;
double rmin;
double rmax;

switch_t bubble_flag;

/* field variables that are defs.h */

field1D *Wt;       // bubble terminal velocity
field1D *Radius;   // bubble radius
field1D *Radiusi;  // inverse bubble radius
field2D *NB;       // bubble size spectrum N(z,r)
field2D *GN;       // bubble breakup freq G(diss(z),r)*N(z,r)
field2D *BR;       // -parent + daughter bubbles term in d/dt(N)
field2D *BDS;      // bubble disolution term (d/dr(D N) in dN/dt 
field1D *Db;       // bubble dissolution speed (m/s)

field2D *NBT1;     // time derivative bubble size spectrum
field2D *NBT2;     // time derivative bubble size spectrum (for AB2)

field1D *BETA;     // void fraction
field1D *BFLUX;    // this is the buoyancy flux  -g/rho0 <rho' w'> due to bubbles

/* Local variable for breakup stuff */

field1D *Radius23;
field1D *TRi;
field1D *TMP;      // temporary memory to use in some functions

field1D *Daughter1;
field1D *Daughter2;
field1D *Percent1;
field1D *Percent2;

field1D *DFLUX;     // tmp variable for the buoyancy flux 

int *BUBBLE_BREAKUP_FLAG;
int *index1;

/* These two functions calculate the Hinze scale - first scalar */

// FIX FIX FIX FIX FIX 

double hinze_critical_radius(double diss)
{
  double r_hinze;

  r_hinze = hinze_critical_radius_multiplyer * pow(diss,-0.4);

  return r_hinze;
}

/* This one does the vector form */

void  hinze_critical_radius_field(field1D *RH, field1D *DISS)
{

  int i;
  int N = DISS->N;
  double *dissd = DISS->data;
  double *rhd = RH->data;

  for (i=0;i<N;i++)
    *rhd++ = hinze_critical_radius_multiplyer * pow(*dissd++,-0.4);

}


/* Solves cubic equation for buble rise velocity 
   a * x^3 + b*x^2 + d = 0                        */

double solve_cubic(double a, double b, double d)
{

  double fx, fpx;
  double x, xold;
  double bp, dp;

  xold = 2.0;
  x = 2.0;

  bp = b/a;  dp = d/a;

  //  printf("For cubic: b' = %lf  d' = %lf\n",bp,dp);
  fx = a*x*x*x + b*x*x + d;

  while (fabs(fx)>1e-10) {
    fpx = 3.0*a*x*x + 2*b*x;
    x -= fx/fpx;
    fx = a*x*x*x + b*x*x + d;
  }  

  //  printf("One root is: %lf \n",x);
  return x;
}




/* Calculates the terminal velocity of the bubbles.   
   Based on Keeling JMR 1993   */

double wterminal_keeling(double r)
{
  double aa,b,d,wtr,wt;
  double tmp;

  /* if bubble radius is big enough then const. rise rate */

  if (r>0.00056) 
    wt = 0.3;  // m/s
  else {
    b=1;
    tmp = 2.0*r/nu;

    aa = sqrt(tmp);
    aa *= 0.0566;
    d = -grav*2.0*r*r/(9.0*nu);

    wtr = solve_cubic(aa,b,d);

    wt = wtr*wtr;
  }

#ifdef DEBUG2
  fprintf(stderr,"For radius: r=%lf (m)  wt=%lf (m/s)\n",r,wt);
#endif  /* DEBUG */

  return wt;

}


void terminal_velocity(field1D *wt, field1D *r)
{

  int i;
  int N = r->N;

  double *wtd = wt->data;
  double *rd = r->data;

  for (i=0;i<N;i++)
    *wtd++ = wterminal_keeling(*rd++);

}


/* Init radius stuff here: Including variables used in breakup calculations */

void init_radius(field1D *r, double rmin, double rmax)
{

  int i, N = r->N;
  double *rd = r->data;
  double *rd23;
  double *rid = Radiusi->data;
  double *Trid = TRi->data;
  double rr;
  const double frac23 = 2.0/3.0;
  //  double T0 = 12.0*surface_tension/rho0;  /* MB part 1 value */
  double T0 = 1.51469396*surface_tension/rho0;     /* Garret et al value */

  dr = (rmax-rmin)/(N-1);
  idr = 1.0/dr;
  
  for (i=0;i<N;i++) {
    rr = rmin + i*dr;
    *rd++ = rr; 
    *rid++ = 1.0/rr;
    *Trid++ = T0/rr;
  }

  rd = r->data;


  rd23 = Radius23->data;

  for (i=0;i<N;i++)
    *rd23++ = pow(*rd++,frac23);

}

/* Calculates void fraction beta = \int r^3 N(r) dr at a particular level */

double calculate_void_fraction_point(field2D *NB, int zlevel)
{

  int i, M = NB->M;

  double *NBd = REF2(NB,zlevel);
  double *rd = Radius->data;
  double r,beta = 0.0;
  
  for (i=0;i<M;i++) {
    r = *rd++;
    beta += r*r*r * (*NBd++);
  }

  beta *= dr*pi*4.0/3.0;;

  return beta;

}


double calculate_void_fraction0(field1D *NR)
{

  int i, M = NB->M;

  double *Nrd = NR->data;
  double *rd = Radius->data;
  double r,beta = 0.0;
  
  for (i=0;i<M;i++) {
    r = *rd++;
    beta += r*r*r * (*Nrd++);
  }

  beta *= dr*pi*4.0/3.0;;

  return beta;

}

/* Calculates void fraction beta = \int r^3 N(r) dr at all levels */

void calculate_void_fraction(const field2D *NB, field1D *BETA)
{

  int i,j, M = NB->M;
  int N = NB->N;

  double *NBd;
  double *rd;
  double r, beta;
  
  for (i=1;i<N-1;i++) {
    rd = Radius->data;
    NBd = REF2(NB,i);
    beta = 0.0;
    for (j=0;j<M;j++) {
       r = *rd++;
       beta += r*r*r * (*NBd++);
    }

    beta *= dr*pi*4.0/3.0;
    DR1(BETA,i) = beta;
  }

  /* Handle Boundary Void Fractions */

  DR1(BETA,0) = DR1(BETA,1);
  DR1(BETA,N-1) = DR1(BETA,N-2);

}


/* This function calculates the energy loss due to bubbles breaking up */

double internal_energy_factor;

void calculate_bubble_breakup_energy_loss(field2D *NB, field1D *LOSS)
{

  int i,j, M = NB->M;
  int N = NB->N;

  double *NBd;
  double *rd;
  double r, loss;
  
  for (i=0;i<N;i++) {
    rd = Radius->data;
    NBd = REF2(NB,i);
    loss = 0.0;
    for (j=0;j<M;j++) {
       r = *rd++;
       loss += r*r * (*NBd++);
    }

    loss *= dr;
    loss *= 4*pi*surface_tension*internal_energy_factor;

    DR1(LOSS,i) = loss;

  }

}


/* Calculates -g/rho0 * <w'rho'> at a particular level */



void calculate_diffusive_buoyancy_flux(field1D *WRHOD, const field2D *NB, const field1D *KV)
{

  int i, N = NB->N;

  const double *betadp = BETA->data+2;
  const double *betadm = BETA->data;
  const double *kvd = KV->data+1;
  const double *kvdm = KV->data;
  double *wrhod = WRHOD->data+1;

  calculate_void_fraction(NB, BETA);

  for (i=1;i<N-1;i++) {
    *wrhod++ = -grav * 0.5*(*kvd++ + *kvdm++) * (*betadp++ - *betadm++) * idz ;
  }

  DR1(WRHOD,0) = DR1(WRHOD,1);
  DR1(WRHOD,N-1) = DR1(WRHOD,N-2);


}



void calculate_buoyancy_flux(field1D *WRHOB, const field2D *NB)
{

  int i,j, M = NB->M;
  int N = NB->N;

  double *NBd;
  const double *wtd;
  const double *rd;
  double r,wrho;

  for (i=1;i<N-1;i++) {
    wtd = Wt->data;
    rd = Radius->data;
    NBd = REF2(NB,i);
    wrho = 0.0;
    for (j=0;j<M;j++) {
       r = *rd++;
       wrho += r*r*r * (*wtd++) * (*NBd++);
    }

    wrho *= dr;
    wrho *= (4.0/3.0)*pi*grav;   /*  /rho0; */
    DR1(WRHOB,i) = wrho;
  }

  /* Handle Boundary Buoyancy Fluxes */

  DR1(WRHOB,0) = DR1(WRHOB,1);
  DR1(WRHOB,N-1) = DR1(WRHOB,N-2);

}



/* This is a no-flux bubble bottom boundary condition */

void update_bubble_bc0(field2D *NN)
{

  int i;
  int N = NN->N;  // number of depth points
  int M = NN->M;  // number of bubble radii

  double *nn1d = NN->data;
  double *nn2d = NN->data+M;
  double *nn3d = NN->row[N-1];
  double *nn4d = NN->row[N-2];
 
  /* First do the bototm boundary condition */

  for (i=0;i<M;i++) {
    *nn1d++ = *nn2d++;
    *nn3d++ = *nn4d++;


  }
}


void init_deane_bubble_spectrum(field2D *NB, field1D *r, double beta0, double scale_h)
{

  static const double n0 = 1e+4;
  static const double a0 = 0.0013;  // 1.3 mm

  int i,j;
  int N = NB->N;
  int M = NB->M;

  double *nbd = NB->data;
  double *rd = r->data;
  double *nbtd;

  double aa, tmp, betam, betat, beta, rr;
  double z;

  /* first calculate the basic spectrum shape */

  for (i=0;i<M;i++) {
    rr = *rd++;
    aa = rr/a0;
    if (aa>1.0)
      tmp=pow(aa,-4.5);
    else {
      if (rr>2e-4) 
	tmp=pow(aa,-2.5);
      else
        tmp = 0.0;
    }
    *nbd++ = n0 * tmp;
  }
  /* implied is that N(M) = 0 - implied +r boundary condition */

  /* next calculate this void fraction */

  betat = calculate_void_fraction_point(NB, 0);

  /* then loop through all the void fractions */

  nbd = NB->data;
  nbtd = NB->data;

  depth = dz * (nz-1);

  for (j=1;j<N-1;j++) {
    z = depth - j*dz;
    beta = beta0 * exp(-z/scale_h);
    betam = beta / betat;
    nbd = NB->data;
    nbtd = REF2(NB,j);
    for (i=0;i<M;i++) 
      *nbtd++ = betam * (*nbd++);
  }

  update_bubble_bc0(NB);


}


void init_bubble_breakup_vars(int M)
{

  int i, dilow;
  double  r, rs;
  double inew, ilow, frac;

  DR1(Daughter1,1) = 0;
  DR1(Daughter2,1) = 0;
  DR1(Percent1,1) = 0.0;
  DR1(Percent2,1) = 0.0;


  for (i=1;i<M;i++) {
    r = rmin+i*dr;
    rs = r/alphas;
    inew = (rs-rmin)/dr;
    frac = modf(inew,&ilow);

    if (ilow<0) {
      ilow = 0;
      frac = 0.0;
    }

    dilow = (int ) ilow;

    DR1(Daughter1,i) = dilow;
    DR1(Daughter2,i) = dilow+1;
    DR1(Percent1,i) = 1.0-frac;
    DR1(Percent2,i) = frac;
  }


}


int round(double a)
{
  double frac,ip;
  int ii;

  frac = modf(a,&ip);
  ii = (int ) ip;

  if (frac>=0.5) 
    ii++;

  return ii;
}

void init_bubble_breakup_vars2(int M, int *index1)
{

  int j, ind;
  double  r, rs;
  double inew;

  index1[M-1] = 0;

  for (j=0;j<M-1;j++) {
    r = rmin+j*dr;
    rs = r*alphas;
    inew = (rs-rmin)/dr;
    ind = round(inew);
    if (ind>M-1) 
      ind=0;
    index1[j] = ind;
  }


}



void init_bubble_stuff(int zlevel, bubbleinfo *B)
{

  bubble_flag = B->bubble_flag;

  if (B->bubble_flag==ON) {
    nr = B->nr;
    rmin = B->rmin;
    rmax = B->rmax;
  }
  else {    /* this is a kludge to make it work FAST */
    nr = 5;
    rmin = 0.0001;
    rmax = 0.001;
  }

  Radius = allocate_field1D(nr);
  Wt = allocate_field1D(nr);
  Radiusi = allocate_field1D(nr);
  TRi = allocate_field1D(nr);
  Radius23 = allocate_field1D(nr);  

  NB = allocate_field2D(zlevel,nr);
  NBT1 = allocate_field2D(zlevel-2,nr);
  NBT2 = allocate_field2D(zlevel-2,nr);

  GN = allocate_field2D(zlevel,nr);
  BR = allocate_field2D(zlevel,nr);    // total breakup term (-parent + daughter)
  BDS = allocate_field2D(zlevel,nr);   // dissolution term  d/dr(Db * N)
  Db = allocate_field1D(nr);           // allocate dissolution spped
  set_field1D_to_value(Db,-1e-6);

  BETA = allocate_field1D(zlevel);    // allocate void fraction memory
  BFLUX = allocate_field1D(zlevel);   // allocate buoyancy flux memory
  DFLUX = allocate_field1D(zlevel);   // allocate buoyancy flux memory
  DR1(DFLUX,0) = 0.0;                 // bottom flux == 0 always
  set_field1D_to_value(BFLUX,0.0);
  set_field2D_to_value(NB, 0.0);      // initial bubble cond = 0

  init_radius(Radius, rmin, rmax);    // init radius

  terminal_velocity(Wt, Radius);      // init terminal velocity

  init_deane_bubble_spectrum(NB, Radius, B->beta0, B->scale_h); 

  internal_energy_factor = pow(2,1.0/3.0)-1.0;

  TMP = allocate_field1D(nr);

  Daughter1 = allocate_field1D(nr);   // daughter1 location
  Daughter2 = allocate_field1D(nr);   // daughter1 location
  Percent1 = allocate_field1D(nr);    // daughter1 percent
  Percent2 = allocate_field1D(nr);    // daughter1 percent
  init_bubble_breakup_vars(nr);        // init the above 4 variables 

  BUBBLE_BREAKUP_FLAG = (int * ) g_malloc(zlevel*sizeof(int));
  index1 = (int * ) g_malloc(nr*sizeof(int));
  init_bubble_breakup_vars2(nr, index1);        // init the above 4 variables 
}


void deallocate_bubble_fields()
{

  deallocate_field1D(Radius);
  deallocate_field1D(Radiusi);
  deallocate_field1D(Radius23);
  deallocate_field1D(TRi);

  deallocate_field1D(Wt);
  deallocate_field1D(Db);
  deallocate_field1D(BETA);
  deallocate_field1D(BFLUX);
  deallocate_field1D(DFLUX);

  deallocate_field2D(NB);
  deallocate_field2D(NBT1);
  deallocate_field2D(NBT2);


  deallocate_field2D(GN); 
  deallocate_field2D(BR);
  deallocate_field2D(BDS);

  deallocate_field1D(Daughter1);
  deallocate_field1D(Daughter2);
  deallocate_field1D(Percent1);
  deallocate_field1D(Percent2);

  g_free(BUBBLE_BREAKUP_FLAG);
  g_free(index1);

}

/* returns the breakup probability of a bubble of radiu r
   and TKE dissipation diss, based on Lasheras JFM paper */
   
const double Kg = 0.15749013123686; // 0.25 * 2^(1/3) / 2
const double bb = 8.2;


/* returns the breakup probability of a bubble of radiu r
   and TKE dissipation diss, based on Lasheras JFM paper 
   -- this does things in vector format and returns the
   2-D field probability G  */


void calculate_parent_breakup_term(field2D *GN, const field1D *R, const field1D *DISS)
{
  int i,j;
  int M = R->N;
  int N = DISS->N;

  double *rd, *r23d;
  double *gnd = GN->data+M;
  double *nbd = NB->data+M;
  double *rid, *Trid;
  double r;
  //  static double T0 = 12.0*surface_tension/rho0;      /* MB part 1 et al. say use 12.0 */
  double T0 = 1.51469396*surface_tension/rho0;    /* Garret quote hinze scale const c=0.36 */
  double diss, diss23, tmp1, tmp2;
  const double frac23 = 2.0/3.0;


  for (i=1;i<N-1;i++) {   // don't have to start at bottom
                        // or go all the way to the top
    diss = DR1(DISS,i);
    diss23 = pow(diss,frac23);

    /* First check to make sure >0 */

    tmp1 = bb*pow(diss*rmin,frac23) - T0/rmin;
    tmp2 = bb*pow(diss*rmax,frac23) - T0/rmax;

    if (tmp2 < 0) {  /* then all are zero */
      BUBBLE_BREAKUP_FLAG[i] = OFF;
      /*  for (j=0;j<M;j++) 
       *gnd++ = 0.0;  */
    }
    else{
      BUBBLE_BREAKUP_FLAG[i] = ON;
#ifdef DEBUG2
      fprintf(stderr,"** bubble.c: Bubble breaking on at level i=%d\n",i);
#endif /* DEBUG2 */

      rd = R->data;
      r23d = Radius23->data;
      rid = Radiusi->data;
      Trid = TRi->data;

    /* iterate through the radii */

      for (j=0;j<M;j++) {
	r = *rd++;
	tmp1 = bb*diss23*(*r23d++) - *Trid++;
	if (tmp1<0) {
	  *gnd++ = 0.0;
          nbd++;
          rid++;
	}
        else
	  *gnd++ = Kg * sqrt( tmp1 ) * (*rid++) * (*nbd++);
      }
    }

  }

}



void calculate_bubble_breakup_terms(field2D *BR, const field1D *DISS)
{

  int i,j;
  int N = BR->N;
  int M = BR->M;
  int d1, d2;
  double breakup, p1 ,p2;
  const double two43 = 2.51984209978975;

  double *gnd, *brd, *tmpd;
  double *dd1, *dd2, *pp1, *pp2;

  /* calculate the parent breakup terms */

  calculate_parent_breakup_term(GN, Radius, DISS);

  for (i=1;i<N-1;i++) {
    gnd = REF2(GN,i)+1;
    brd = REF2(BR,i);
    dd1 = Daughter1->data+1;
    dd2 = Daughter2->data+1;
    pp1 = Percent1->data+1;
    pp2 = Percent2->data+1;
    if (BUBBLE_BREAKUP_FLAG[i]==ON) {
      set_field1D_to_value(TMP,0.0);
      for (j=1;j<M;j++) {
        breakup = two43*(*gnd++);
        d1 = *dd1++;   // DR1(Daughter1,j);  // *dd1++
        d2 = *dd2++;   //DR1(Daughter2,j);  // *dd2++
        p1 = *pp1++;   //DR1(Percent1,j);   // *pp1++
        p2 = *pp2++;   //DR1(Percent2,j);   // *pp2++
        DR1(TMP,d1) += p1*breakup;
        DR1(TMP,d2) += p2*breakup;
      }  /* for (j=1;.... */
      gnd = REF2(GN,i);
      tmpd = TMP->data;
      for (j=0;j<M;j++)
        *brd++ = - (*gnd++) + (*tmpd++);
    } /* if (BUBBLE...) */
    else {
      for (j=0;j<M;j++)
        *brd++ = 0.0;
    } 

  }


}


void calculate_bubble_breakup_terms2(field2D *BR, const field1D *DISS)
{

  int i,j, j1;
  int N = BR->N;
  int M = BR->M;
  double breakup;
  const double two43 = 2.51984209978975;

  double *gnd, *brd, *tmpd;

  /* calculate the parent breakup terms */

  calculate_parent_breakup_term(GN, Radius, DISS);

  for (i=1;i<N-1;i++) {
    gnd = REF2(GN,i);
    brd = REF2(BR,i);
    tmpd = TMP->data;
    if (BUBBLE_BREAKUP_FLAG[i]==ON) {
      for (j=0;j<M-1;j++) {
        j1 = index1[j];
        breakup = gnd[j1];
        *tmpd++  = breakup*two43;
      }  /* for (j=0;.... */
      gnd = REF2(GN,i);
      tmpd = TMP->data;
      for (j=0;j<M;j++)
        *brd++ = - (*gnd++) + (*tmpd++);
    } /* if (BUBBLE...) */
    /*    else {
      for (j=0;j<M;j++)
        *brd++ = 0.0;
    } 
    */
  }


}


void calculate_bubble_dissolution(field2D *BDS, field1D *Kb, field2D *NB)
{

  int i,j;
  int N = BDS->N;
  int M = BDS->M;
  double *bdsd, *nbd;
  double *tmpd, *kbd;
  double tmp, tmp0;

  for (i=1;i<N-1;i++) {
    bdsd = REF2(BDS,i);  // handle the pointer initializations 
    nbd = REF2(NB,i);
    tmpd = TMP->data;
    kbd = Kb->data;
    tmp0 = *kbd++ * (*nbd++);
    for (j=0;j<M-1;j++) {     // first calulate dissolution flux, then grad
      tmp = *kbd++ * (*nbd++);
      *bdsd++ = idr*(tmp-tmp0);
      tmp0 = tmp;
    }
    *bdsd++ = -idr*tmp0;
  }
}



void calculate_bubble_stats(field1D *DISS)
{

#ifdef BUBBLE_BREAKING
  calculate_bubble_breakup_terms2(BR, DISS);
  //  calculate_bubble_breakup_tke_loss(TKELOSS);
#endif /* BUBBLE BREAKING */

#ifdef BUBBLE_DISSOLUTION
  calculate_bubble_dissolution(BDS,Db,NB);
#endif /* BUBBLE DISSOLUTION */


}





void rhs_implicit_bubble_eq(field2D *NBT, const field2D *NB)
{

  int i,j;
  int N = NB->N;
  int M = NB->M;
  int lim;

  //  double Fup, Fdn, diffusion;
  double rise=0.0, breakup = 0.0;
  double dissolution = 0.0;

  double kv;
  double kvm;
  //  double totFup;
  double rr;

  double *nbtd = NBT->data;

  double *nbdp = NB->data+2*M;
  double *nbd = NB->data+M;
  double *nbdm = NB->data+M;
  double *nbdm2;
  double *wtd = Wt->data;
  double *radiusd = Radius->data;
  double *brd = BR->data+M;
  double *bdsd = BDS->data+M;
  double *bsourced = NN_SOURCE->data+M;
  //  double *rhowd = RHOW->data+1;

  /* First do the near bottom w/ centered advection for BC */

  i = 1;

  //  kv = DR1(Kv,i);
  //  totFup = 0.0;

  for (j=0;j<M;j++) {

    rr = *radiusd++;

   /* this is the bubble diffusion stuff */

    //Fup = DR1(Kv,i)*(DR2(NB,i+1,j)-DR2(NB,i,j));

    //    Fup = kv * (*nbdp++ - *nbd);
    //Fdn = 0.0;
    //    totFup += CUBE(rr)*Fup;

    //    diffusion = idz2*(Fup-Fdn);

    /* Bubble rise stuff */
    /* d/dx(wN) ~  d_z (\bar(w N)^z)  */
    /* close to boundary just first order upwind */

    rise = 0.5*idz* (*wtd++) * (*nbd++); // - 0.0;

    /* Bubble disolution */

#ifdef BUBBLE_DISSOLUTION
    dissolution = *bdsd++;
#endif /* BUBBLE_DISSOLUTION */

    /* add it all together */

    *nbtd++ =  - (rise + dissolution);

  }

#ifdef BUBBLE_BREAKING
  if (BUBBLE_BREAKUP_FLAG[i]==ON) {
    nbtd -= M;
    brd = REF2(BR,i);
    for (j=0;j<M;j++) 
      *nbtd ++ += breakup;
  }
#endif /* BUBBLE BREAKING */

  /* deal w/ diffusive buoyancy flux */

  //  DR1(DFLUX,i) = totFup*dr*idz;


  /* ----------------------------------------- */
  /* reset rise-speed & radius pointers  */

  //  totFup = 0.0;
  wtd = Wt->data;
  radiusd = Radius->data;

  i = 2;
  //  kv = DR1(Kv,i);
  //  kvm = DR1(Kv,i-1);

  for (j=0;j<M;j++) {

    rr = *radiusd++;

   /* this is the bubble diffusion stuff */

    //    Fup = kv *(*nbdp - *nbd);
    //    Fdn = kvm *(*nbd - *nbdm);
    //    totFup += CUBE(rr)*Fup;

    //    diffusion = idz2*(Fup-Fdn);

    /* Bubble rise stuff */
    /* d/dx(wN) ~  d_z (\bar(w N)^z)  */

    rise = idz*(*wtd++)*(2.0*(*nbdp++) + 3.0*(*nbd++) -6.0*(*nbdm++))/6.0;

    /* Bubble disolution */

#ifdef BUBBLE_DISSOLUTION
    dissolution = *bdsd++;
#endif /* BUBBLE_DISSOLUTION */

    /* add it all together */

    *nbtd++ = -(rise + dissolution);

  }

#ifdef BUBBLE_BREAKING
  if (BUBBLE_BREAKUP_FLAG[i]==ON) {
    nbtd -= M;
    brd = REF2(BR,i);
    for (j=0;j<M;j++) 
      *nbtd ++ += breakup;
  }
#endif /* BUBBLE BREAKING */

  /* deal w/ diffusive buoyancy flux */

  //  DR1(DFLUX,i) = totFup*dr*idz;

  /* ----------------------- */

  nbdm2 = NB->data+M;  // this is needed for 3rd order upwind

  for (i=3;i<N-1;i++) {
    //    totFup = 0.0;
    wtd = Wt->data;     /* reset the rise velocity */
    radiusd = Radius->data;
    //    kv = DR1(Kv,i);
    //    kvm = DR1(Kv,i-1);

    for (j=0;j<M;j++) {
  
      rr = *radiusd++;

      /* this is the bubble diffusion stuff */

      //      Fup = kv *(*nbdp - *nbd);
      //      Fdn = kvm *(*nbd - *nbdm);
      //      totFup += CUBE(rr)*Fup;

      //      diffusion = idz2*(Fup-Fdn);

      /* Bubble rise stuff */
      /* d/dx(wN) ~  d_z (\bar(w N)^z)  */

      rise = idz*(*wtd++)*(2.0*(*nbdp++) + 3.0*(*nbd++) -6.0*(*nbdm++) + *nbdm2++)/6.0;
      
    /* Bubble disolution */

#ifdef BUBBLE_DISSOLUTION
    dissolution = *bdsd++;
#endif /* BUBBLE_DISSOLUTION */

 
      /* add it all together */

    *nbtd++ = - (rise + dissolution);
    
    }   /* for (j=0.... */

    /* Bubble breakup (-parent + daughters) term */

#ifdef BUBBLE_BREAKING
    if (BUBBLE_BREAKUP_FLAG[i]==ON) {
      nbtd -= M;
      brd = REF2(BR,i);
      for (j=0;j<M;j++) 
	*nbtd ++ += breakup;
    }
#endif /* BUBBLE BREAKING */

  /* deal w/ diffusive buoyancy flux */

    //  DR1(DFLUX,i) = totFup*dr*idz;

  }  /* for (i=3;......) */

  /* Here take care of plunging wave bubble source */

  if ((turb_source==PLUNGING)&&(wave_breaking==ON)) {
    nbtd = NBT->data;
    bsourced = NN_SOURCE->data+M;
    lim = M*(N-2);
    for (i=0;i<lim;i++)
      *nbtd++ += *bsourced++;
  }



  /* Here do the buoyancy flux calculations */
  /*
  for (i=1;i<N-1;i++) {
    *rhowd++ = -grav*(2.0/3.0)*pi*(DR1(DFLUX,i)+DR1(DFLUX,i-1));
  }
  DR1(RHOW,0) = 0.0;
  DR1(RHOW,N-1) = DR1(RHOW,N-2);
  */
}      



