/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*    
      density.c  - density module to turbulence model 
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <glib.h>

#include "tri.h"
#include "density.h"


/* Extern density field variables */

switch_t density_flag = OFF;
bc_t top_bc;
double bcval;

field1D *WRHO;        // water density - 1000kg/m^3
field1D *WRHOFLUX;    // buoyancy flux due to the water density
field1D *KRHOV;

field1D *N2;          // buoyancy frequency squared
field1D *S2;           // shear squared
field1D *alphaN;      // stability parameters
field1D *alphaM;      

/* Non-extern mean flow variables */

field1D *DIDZ1;
field1D *DIDZ2;
double  surface_buoyancy_flux;

/* implicit solver variables */

double *rhsr;
double *dd;
double *ld;
double *ud;   
double *workr;



/* Density Solver -> since DIDZ is constant or not */

void implicit_solve_density_euler(field1D *RHOW, const field1D *DIDZ, 
                                  const field1D *Kv, double dt)
{

  int i, ok;
  int N = RHOW->N;

  double *rhowd = RHOW->data+1;

  double *rhofluxd = WRHOFLUX->data+1;

  double rho, rhop;

  const double *didzd = DIDZ->data;

  const double *kvd = Kv->data;

  double kv, kvm, total;
  double Fp, Fm, tmp;
  double rho_diffusion;

  double hdt = 0.5*dt;

  /* First calculate the RHS stuff */
  /* do the first grid minus flux before looping */

  rho = *rhowd++;

  kvm = *kvd++;
  Fm = kvm * (rho - DR1(RHOW,0));

  for (i=1;i<N-1;i++) {

    rhop = *rhowd++;
    kv = *kvd++;

    Fp = kv * ( rhop - rho);

    *rhofluxd++ = grav*idz*0.5*(Fp+Fm)/rho0;

    rho_diffusion = idz2*(Fp-Fm);
    rhsr[i-1] = rho + dt*(0.5*rho_diffusion + *didzd++);

    Fm = Fp;
    rho = rhop;

  }   /* for (i=0.... */

  /* Now the tridiagonal matrix is set up */
  /* i = 0 -> here we apply no-slip bottom bc */

  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;

  dd[0] = 1 + hdt*idz2*(kv);   // bed is no-flux bc.
  ud[0] = -hdt*kv*idz2;
  ld[0] = 0.0;
  kvm = kv;

  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    dd[i] = 1 + hdt*idz2*(kvm+kv);
    ud[i] = -hdt*kv*idz2;
    ld[i] = -hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, here deal with top flux BC */

  i = N-3;
  kv = *kvd++;
  ld[i] = -hdt*kvm*idz2;
  ud[i] = 0.0;
  switch (top_bc) {
  case FLUX: total = kvm;
     rhsr[i] += idz*hdt*bcval;
     break;
  case FIXED: total = kvm+kv;
     rhsr[i] += hdt*kv*idz2*bcval;   // this is for fixed bc
  }

  dd[i] = 1.0 + hdt*idz2*(total);

  /* Now do the implicit solving */

  rhowd = RHOW->data+1;   // reset the rho pointer

  ok = tridiagonal(dd, ld, ud, rhsr, rhowd, workr, N-2);
  if (ok != 0) {
    if (ok==-1)
      error_message("** Critical Error:  Tridiagonal call for density failed - diag[0]==0.0!");

    printf("** Bad bet=0.0 at j=%d\n",ok);
    error_message("** Critical Error:  Tridiagonal call for density failed!");
  }

  /* Next is rho boundary conditions */
  /* Bottom BC is no density flux near bed */

  DR1(RHOW,0) = DR1(RHOW,1);

  /* Surface BC */

  switch(top_bc) {
  case FIXED:  DR1(RHOW,N-1) = bcval;
               break;
  case FLUX:   DR1(RHOW,N-1) = DR1(RHOW,N-2) + bcval*dz/kv;
  }

  DR1(WRHOFLUX,N-1) = DR1(WRHOFLUX,N-2);
  DR1(WRHOFLUX,0) = DR1(WRHOFLUX,1);


}




void implicit_solve_densityAB2(field1D *RHOW, const field1D *DIDZ1, const field1D *DIDZ2,
                                  const field1D *Kv, double dt)
{

  int i, ok;
  int N = RHOW->N;

  double *rhowd = RHOW->data+1;

  double *rhofluxd = WRHOFLUX->data+1;

  double rho, rhop;

  const double *didzd1 = DIDZ1->data;
  const double *didzd2 = DIDZ2->data;

  const double *kvd = Kv->data;

  double kv, kvm, total;
  double Fp, Fm, tmp;
  double rho_diffusion;

  double hdt = 0.5*dt;

  /* First calculate the RHS stuff */
  /* do the first grid minus flux before looping */

  rho = *rhowd++;

  kvm = *kvd++;
  Fm = kvm * (rho - DR1(RHOW,0));

  for (i=1;i<N-1;i++) {

    rhop = *rhowd++;
    kv = *kvd++;

    Fp = kv * ( rhop - rho);

    *rhofluxd++ = grav*idz*0.5*(Fp+Fm)/rho0;

    rho_diffusion = idz2*(Fp-Fm);
    rhsr[i-1] = rho + hdt*(rho_diffusion + 3.0*(*didzd1++) - (*didzd2++));

    Fm = Fp;
    rho = rhop;

  }   /* for (i=0.... */

  /* Now the tridiagonal matrix is set up */
  /* i = 0 -> here we apply no-slip bottom bc */

  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;

  dd[0] = 1 + hdt*idz2*(kv);   // bed is no-flux bc.
  ud[0] = -hdt*kv*idz2;
  ld[0] = 0.0;
  kvm = kv;

  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    dd[i] = 1 + hdt*idz2*(kvm+kv);
    ud[i] = -hdt*kv*idz2;
    ld[i] = -hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, here deal with top flux BC */

  i = N-3;
  kv = *kvd++;
  ld[i] = -hdt*kvm*idz2;
  ud[i] = 0.0;
  switch (top_bc) {
  case FLUX: total = kvm;
     rhsr[i] += idz*hdt*bcval;
     break;
  case FIXED: total = kvm+kv;
     rhsr[i] += hdt*kv*idz2*bcval;   // this is for fixed bc
  }

  dd[i] = 1.0 + hdt*idz2*(total);

  /* Now do the implicit solving */

  rhowd = RHOW->data+1;   // reset the rho pointer

  ok = tridiagonal(dd, ld, ud, rhsr, rhowd, workr, N-2);

  if (ok != 0) {
    if (ok==-1)
      error_message("** Critical Error:  Tridiagonal call for density failed - diag[0]==0.0!");

    printf("** Bad bet=0.0 at j=%d\n",ok);
    error_message("** Critical Error:  Tridiagonal call for density failed!");
  }

  /* Next is rho boundary conditions */
  /* Bottom BC is no density flux near bed */

  DR1(RHOW,0) = DR1(RHOW,1);

  /* Surface BC */

  switch(top_bc) {
  case FIXED:  DR1(RHOW,N-1) = bcval;
               break;
  case FLUX:   DR1(RHOW,N-1) = DR1(RHOW,N-2) + bcval*dz/kv;
  }

  DR1(WRHOFLUX,N-1) = DR1(WRHOFLUX,N-2);
  DR1(WRHOFLUX,0) = DR1(WRHOFLUX,1);


}


#define CANUTO_A

void calculate_density_stats(field1D *DIDZ, field1D *KRHOV, field1D *KUV)
{

  int N = KRHOV->N;
  int i;

  const static double iPr=1.4;    // this is -g/(rho0 * Pr)

  const static double Cmu0 = 0.5562;
  const static double Cmu0_4 = 9.570264439923362e-02;  

  double *wrhod = WRHO->data;

  double *krhovd = KRHOV->data;
  double *kuvd = KUV->data;
  double *bd = BB->data;
  double *dissd = DISS->data;
  double *N2d = N2->data;
  double *S2d = S2->data;
  double *ud = UU->data;
  double *vd = VV->data;

  double rhom, rhop;
  double Nsqr, Ssqr;
  double km, kp;
  double em, ep;
  double up, um, vp, vm;
  double alpha_N, alpha_M;
  double num, den, fac;

  double kuv;
  double Cmu;
  double Cmu_prime;


  rhom = *wrhod++;
  km = *bd++;
  em = *dissd++;
  um = *ud++;
  vm = *vd++;

  for (i=0;i<N;i++) {
 
    kp = *bd++;
    ep = *dissd++;
    rhop = *wrhod++;
    up = *ud++;
    vp = *vd++;

    Ssqr = idz2 * (SQR((up-um)) + SQR((vp-vm)));
    *S2d++ = Ssqr;

    fac = 0.25* SQR( (kp/ep + km/em));   //((kp+km)/(ep+em)) );

    alpha_M = fac * Ssqr;     //  (k^2/epsilon^2) * S^2

    Nsqr = -grav * idz* (rhop-rhom)/rho0;
    *N2d++ = Nsqr;

    alpha_N =  fac * Nsqr;    //  (k^2/epsilon^2) * N^2

    if ((alpha_N < 0)) {
      //      printf("** Warning, alpha_N < 0, alpha_N = %g at i=%d and time=%g (sec)\n",alpha_N,i,global_time);
      //      printf("** Nsqr=%g, kp=%g, km=%g,  ep=%g, em=%g \n",Nsqr,kp,km,ep,em);
      alpha_N = 0;

    }

    if (alpha_N>1) 
      alpha_N=1;


    kuv = *kuvd;

#ifdef GALPERIN
    Cmu = (Cmu0 + 2.18*alpha_N)/(1+20.40*alpha_N + 53.12*alpha_N*alpha_N);
    Cmu_prime = 0.6985/(1+17.34*alpha_N);
    *kuvd++ = (Cmu/Cmu0) * kuv;
    *krhovd++ = (Cmu_prime/Cmu0) * kuv;
#endif


#ifdef CANUTO_A
    num = 0.1070 + 0.01741*alpha_N - 0.00012*alpha_M;
    den = 1 + 0.2555*alpha_N + 0.02872*alpha_M + 0.008677*SQR(alpha_N) + 0.005222*alpha_N*alpha_M 
      - 0.0000337*SQR(alpha_M);

    if (num<1e-4) 
      num = 1e-4;

    if (den<=1e-3)
      den = 1e-3;

    Cmu = num/den;

    if (Cmu<1e-3)
      Cmu = 1e-3;

    if (Cmu>5)
      Cmu = 5;

    num = 0.1120 + 0.004519*alpha_N - 0.00088*alpha_M;

    if (num<1e-4) 
      num = 1e-4;

    Cmu_prime = num/den;

    if (Cmu_prime<1e-3)
      Cmu_prime = 1e-3;

    if (Cmu_prime>5)
      Cmu_prime = 5;

    *kuvd++ = (Cmu/Cmu0_4) * kuv;
    *krhovd++ = (Cmu_prime/Cmu0_4) * kuv;

#endif /* CANUTO_A */




    km = kp;
    em = ep;
    rhom = rhop;
    um = up;
    vm = vp;


  }

  // do nothing right now with light heating

}



void  init_density_with_BVfreq(field1D *WRHO, double BVfreq)
{
  const double wrho0 = 1028;  /* kg/m^3 */
 
  int N = WRHO->N;
  int i;
  double factor;

  DR1(WRHO,N-1) = wrho0;
  factor = dz*wrho0*BVfreq*BVfreq/grav;

  for (i=N-2;i>=0;i--) {
    DR1(WRHO,i) = DR1(WRHO,i+1) + factor;
    DR1(N2,i) = BVfreq*BVfreq;
  }


}

void allocate_density_variables(int nz, densityinfo *D)
{
  double tmp;
 
  density_flag = D->density_flag;  // set whether density stuff is on or off
  top_bc = D->top_bc;
  bcval = D->bcval;
  
  /* allocate the memory to be used */

  WRHO  = allocate_field1D(nz);
  WRHOFLUX = allocate_field1D(nz);
  KRHOV = allocate_field1D(nz-1);
  DIDZ1 = allocate_field1D(nz);
  DIDZ2 = allocate_field1D(nz);
  N2 = allocate_field1D(nz-1);      // buoyancy frequency
  S2 = allocate_field1D(nz-1);      // shear squared
  alphaN = allocate_field1D(nz-1);  // is this the right size??
  alphaM = allocate_field1D(nz-1);  

  set_field1D_to_value(WRHO,0.0);
  set_field1D_to_value(WRHOFLUX,0.0);
  set_field1D_to_value(DIDZ1,0.0);
  set_field1D_to_value(DIDZ2,0.0);

  /* initialize the density field */

  init_density_with_BVfreq(WRHO,D->BVfreq);

  dd = (double * ) g_malloc(sizeof(double)*(nz-2));
  ud = (double * ) g_malloc(sizeof(double)*(nz-2));
  ld = (double * ) g_malloc(sizeof(double)*(nz-2));
  rhsr = (double * ) g_malloc(sizeof(double)*(nz-2));
  workr = (double * ) g_malloc(sizeof(double)*(nz-2));

  /* Here set up the initial condition w/ the BVfreq stratification */


}


void deallocate_density_variables()
{

  deallocate_field1D(WRHO);
  deallocate_field1D(WRHOFLUX);
  deallocate_field1D(DIDZ1);
  deallocate_field1D(DIDZ2);
  deallocate_field1D(alphaN);
  deallocate_field1D(alphaM);
  deallocate_field1D(KRHOV);
  deallocate_field1D(N2);
  deallocate_field1D(S2);

  g_free(dd);
  g_free(ud);
  g_free(ld);
  g_free(workr);
  g_free(rhsr);

}


