/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*    
      kepsilon.c  - turbulence & bubble model
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <glib.h>
#include <time.h>

#include "defs.h"

#if THREAD_YES==1      /* this must be after the "defs.h" include */
#include <pthread.h>
#endif /* THREAD_YES */


#include "time_stepping.h"
#include "pinput.h"
#include "input.h"
#include "output.h"
#include "cmdline.h"
#include "gui.h"
#include "bubble.h"
#include "misc.h"
#include "wave.h"
#include "mflow.h"
#include "tke.h"
#include "sediment.h"
#include "implicit.h"
#include "q2l.h"



/* Here the model domain size parameters are defines */


int nz;
double idz;       // This is in units of 1/m, grid spacing =
double idz2;
double dz;
double depth;
double edt;       // this is the time step - don't want to call it "dt"

const double grav = 9.81;           // m/s^2 
const double nu = 1.2e-6;           // units and size??
const double rho0 = 1025;           // units kg/m^3
const double pi = 3.14159265358979; 

double global_time;  
 


void error_message(char *s)
{

  fprintf(stderr,"%s - at time = %g (sec)\n",s,global_time);
  kepsilon_exit(-1);

}

/* ---------- INIT ROUTINES --------------*/ 


/* This function should be called in the initialization phase to zero out
   all the matrices  */

void initialize_all_variables(inputinfo *I)
{

   global_time = 0.0;

   nz = I->nz;   
   dz = I->dz; 
   idz = 1.0/dz;
   idz2 = idz*idz; 
   depth = I->depth;


   /* TKE initialization called first */

   init_tke_fields(I->nz,I->z0b,I->z0s);

   /* Bubble initializations */

   init_bubble_stuff(I->nz, I->B);

   /* initialize wave stuff, called AFTER bubble stuff */

   init_wave_parameters(I->nz, I->B->nr, I->W, I->B->scale_h, I->T->dt, I->M, I->z0s);

   allocate_mflow_variables(I->nz, I->M);

  /* load in the TKE IC */

   switch(I->tke_ic) {
   case -1:   break;                    // zero out things   
   case 0:    //set_const_tke(BB,atof(I->tke_ic_file->str)); 
              break;
   case 1:    load_tke_ic(I->tke_ic_file->str, BB);
              break;
   default:   error_message("Wrong TKE IC flag in init file");
   }

   calculate_tke_stats(QQ, DISS, KQV, KUV, BB, LL);

   /* sediment  module initializations */

   edt = I->T->dt;

   allocate_sediment_variables(I->nz,I->SED);

   /* density  module initializations */

   allocate_density_variables(I->nz, I->D);
   check_dissipation_stratified(DISS, N2);  // adjust dissipation for N2

   /* implicit solver initializations */

   init_implicit_solver(I->nz,I->B->nr, I->B->bubble_flag);


}


/* -- Free up allocated memory for each module -- */

void  deallocate_all_fields()
{

  deallocate_tke_fields();
  deallocate_bubble_fields();
  deallocate_mflow_variables();
  deallocate_wave_variables();
  deallocate_sediment_variables();

}




/* ------- Time Integration Routines ------------ */



void level1_integrate_CB(timing *T, cmdline *Cmd, oflags1 *level1)
{
  int i;

  for (i=0;i<T->level1_iter;i++) {
    time_stepAB2_CB(T);
    }                       //  end of level 1 loop:  for(i=0.....  

  /* first calculate the void fraction if bubble active  */

  if (bubble_flag == ON) 
    calculate_void_fraction(NB, BETA);

  if (sediment_flag == ON)
    calculate_sediment_fluxes(SFLUX_U,SFLUX_V);

  global_time = T->current_time;
  level1_output(level1);           // level1 output stuff

  if (Cmd->diagnostic==1) {
    stability_report(KUV,T,dz);
  }

}

void level1_integrate_kepsilon(timing *T, cmdline *Cmd, oflags1 *level1)
{
  int i;

  for (i=0;i<T->level1_iter;i++) {
    time_stepAB2_kepsilon(T);
    }                       //  end of level 1 loop:  for(i=0.....  

  /* first calculate the void fraction if bubble active  */

  if (bubble_flag == ON) 
    calculate_void_fraction(NB, BETA);

  if (sediment_flag == ON)
    calculate_sediment_fluxes(SFLUX_U,SFLUX_V);

  global_time = T->current_time;
  level1_output(level1);           // level1 output stuff

  if (Cmd->diagnostic==1) {
    stability_report(KUV,T,dz);
  }

}



void level1_integrate_q2l(timing *T, cmdline *Cmd, oflags1 *level1)
{
  int i;

  for (i=0;i<T->level1_iter;i++) {
    time_stepAB2_q2l(T);
    }                       //  end of level 1 loop:  for(i=0.....  

  /* first calculate the void fraction if bubble active  */

  if (bubble_flag == ON) 
    calculate_void_fraction(NB, BETA);

  if (sediment_flag == ON)
    calculate_sediment_fluxes(SFLUX_U,SFLUX_V);

  global_time = T->current_time;
  level1_output(level1);           // level1 output stuff

  if (Cmd->diagnostic==1) {
    stability_report(KUV,T,dz);
  }

}


void level3_integrate_CB(timing *T, cmdline *Cmd, outputflags *A)
{
  int j,k;

  T->current_time = 0.0;

  /* First do any level 0 output before the first time step */

  calculate_void_fraction(NB, BETA);  // first calc void fraction 
  level1_output(A->F0);

  /* Now do the first time step */

  time_step_zero_CB(T);      // do the first time step 

  /* iterate through the rest of the time steps write kinetic energy if verbose*/
  /* This is the heart of the code */

  for (j=0;j<T->level3_iter;j++) {

    for (k=0;k<T->level2_iter;k++) {

      level1_integrate_CB(T, Cmd, A->F1);

     }         //  end of level 2 loop: for(k=0......

    level1_output(A->F2);      // level2 output stuff

    if (Cmd->diagnostic==2) {
      stability_report(KUV,T,dz);
    }

  }                                  // end of loop: for(j=0....

  level1_output(A->F3);

  if (Cmd->diagnostic==3) {
    stability_report(KQV,T,dz);
  }

}


void level3_integrate_kepsilon(timing *T, cmdline *Cmd, outputflags *A)
{
  int j,k;

  T->current_time = 0.0;

  /* First do any level 0 output before the first time step */

  //  calculate_void_fraction(NB, BETA);  // first calc void fraction 
  level1_output(A->F0);

  /* Now do the first time step */

  time_step_zero_kepsilon(T);      // do the first time step 

  global_time = T->current_time;

  /* iterate through the rest of the time steps write kinetic energy if verbose*/
  /* This is the heart of the code */

  for (j=0;j<T->level3_iter;j++) {

    for (k=0;k<T->level2_iter;k++) {

      level1_integrate_kepsilon(T, Cmd, A->F1);

     }         //  end of level 2 loop: for(k=0......

    level1_output(A->F2);      // level2 output stuff

    if (Cmd->diagnostic==2) {
      stability_report(KUV,T,dz);
    }

  }                                  // end of loop: for(j=0....

  level1_output(A->F3);

  if (Cmd->diagnostic==3) {
    stability_report(KQV,T,dz);
  }

}


void level3_integrate_q2l(timing *T, cmdline *Cmd, outputflags *A)
{
  int j,k;

  T->current_time = 0.0;

  /* First do any level 0 output before the first time step */

  calculate_void_fraction(NB, BETA);  // first calc void fraction 
  level1_output(A->F0);


  /* Now do the first time step */

  time_step_zero_q2l(T);      // do the first time step 

  /* iterate through the rest of the time steps write kinetic energy if verbose*/
  /* This is the heart of the code */

  for (j=0;j<T->level3_iter;j++) {

    for (k=0;k<T->level2_iter;k++) {

      level1_integrate_q2l(T, Cmd, A->F1);

     }         //  end of level 2 loop: for(k=0......

    level1_output(A->F2);      // level2 output stuff

    if (Cmd->diagnostic==2) {
      stability_report(KUV,T,dz);
    }

  }                                  // end of loop: for(j=0....

  level1_output(A->F3);

  if (Cmd->diagnostic==3) {
    stability_report(KQV,T,dz);
  }

}


#if THREAD_YES==1
pthread_t gui_thread;
pthread_t main_thread;
void *gui_thread_result;
switch_t gui_flag = OFF;
#endif /* THREAD_YES */

/* Generic exit function that cleans up the various threads before 
   exiting main.    This doesn't work yet           */

void kepsilon_exit(int exitcode)
{

#if THREAD_YES==1

  if (gui_flag == ON) {
    pthread_cancel(gui_thread);     // first the thread must be signalled to die 
    pthread_join(gui_thread,&gui_thread_result);     // Then we have to collect the thread 
  }

#endif /* THREAD_YES */

  exit(exitcode);

}

int main(int argc,char *argv[])
{
  int status;
  inputinfo *I;
  timing *T;
  cmdline *Cmd;
  outputflags *A;

#if THREAD_YES==1
  gtk_init(&argc,&argv);   // init the GUI interface 

  main_thread = pthread_self();
#endif /*THREAD YES */

  Cmd = process_args2(argc,argv);

#if THREAD_YES==0
  if (Cmd->gui) {
    fprintf(stderr,"** Warning: No GUI output possible w/ no thread capability compiled in!\n");
  }
#endif /* THREAD_YES */

  I = load_init_file(Cmd->init_file);

  A = I->A;
  T = I->T;

  defs_test();    // test to make sure no uncompatible defines are set


  if (Cmd->diagnostic) {
    defs_report();           // report on what model options are set
    stability_report(KQV,T,dz);
    timing_report(T);             // report on the timing variables
    wave_and_forcing_report(I);
  }

  if (Cmd->parse) {
     error_message("Finished Parsing init file\n");
  }


  /* Start the Gui timer */
#if THREAD_YES==1
  if (Cmd->gui) {
    T->msg = Cmd->msg;
    T->message = Cmd->message;
    gui_flag = ON;
    status = pthread_create(&gui_thread, NULL, (void *) gui_start, T);
  }
#endif /* THREAD_YES */

  /* do the integration */

  switch (I->model_type) {
  case KEPSILON:      level3_integrate_kepsilon(T, Cmd, A); 
    break;
  case MELLOR_YAMADA: level3_integrate_q2l(T, Cmd, A);
    break;
  case CRAIG_BANNER:  level3_integrate_CB(T, Cmd, A); 
  }


  /* free all the allocated memory */

  outputflags_destroy(A);
  deallocate_all_fields();   // destroy all TKE & Bubble fields stuff

#if THREAD_YES==1
  if (Cmd->gui) 
    pthread_join(gui_thread,&gui_thread_result);
#endif /* THREAD_YES */

  return 0;

}


  




