/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/* --  output.c:  handles output for model */



#include <stdio.h>
#include <stdlib.h>

#include "output.h"

/* this function is in misc.c */

double  diffusive_stability_parameter(field1D *KV, double dt, double dz);


FILE*  open_file_for_writing(char *s)
{
  FILE *fp;

  if ((fp=fopen(s,"w"))==NULL) {
    fprintf(stderr,"Cannot open file for writing: %s\n",s);
    exit(-1);
  }

  return fp;
}
   

/* 
   This function writes a longshore array in ascii in one row.
   The array is written at cross-shore location i - with checking

*/

void  write_field1D_ascii(FILE *fout, field1D *F)
{

     int i;
     int N = F->N;

     for (i=0;i<N;i++)
       fprintf(fout,"%e ",DR1(F,i));

     fprintf(fout,"\n");

}

/* Same as above but in binary */

void  write_field1D_binary(FILE *fout, field1D *F)
{

     size_t size;

     size = sizeof(double);

     fwrite(F->data,size,F->N,fout);

}


double calculate_total_tke(field1D *BB, double dz)
{

  int i;
  int N = BB->N;
  double tot_tke = 0.0;

  const double *bbd = BB->data;
  const double *bbdp = BB->data+1;

  for (i=0;i<N-1;i++) 
    tot_tke += 0.5*(*bbd++ + *bbdp++);
    
  tot_tke *= dz;

  return tot_tke;
}


double calculate_total_pe(field1D *BETA, double dz)
{

  int i;
  int N = BB->N;
  double tot_pe = 0.0;
  double z0, zp;

  const double *betad = BETA->data;
  const double *betadp = BETA->data+1;

  for (i=0;i<N-1;i++) {
    z0 = i*dz;
    zp = z0+dz; 
    tot_pe += 0.5*(z0 * *betad++ + zp * *betadp++);
  }    
  tot_pe *= grav*dz;

  return tot_pe;

}


/* --- oflags1 & level1 output routines --- */

void output_point(point *P)
{

  double pt;
  output_options_t *O = P->OP;

  /* calculate the point variable */

  switch (P->var) {
  case TOTTKE_M:  pt = calculate_total_tke(BB,dz);
    break;
  case TOTPE_M:  pt = calculate_total_pe(BETA,dz);
    break;
  case TOTE_M:  pt = calculate_total_pe(BETA,dz) + calculate_total_tke(BB,dz);
    break;
  case STABILITY_M: pt = diffusive_stability_parameter(KUV, edt, dz);
    break;
  case TKE_M:
  case U_M:
  case V_M:
  case DISS_M:
  case BETA_M:
  case KQV_M:
  case KUV_M:
  case PROD_M:
  case TKEDIFF_M:
  case DBDT_M:
  case UW_M:
  case VW_M:
  case Q2L_M:
  case SEDIMENT_M:
  case SEDIMENT_FLUX_U_M:
  case SEDIMENT_FLUX_V_M:
  case WRHO_M:
  case WRHOFLUX_M:
  case KRHOV_M:
  case BFLUX_M:    pt = DR1(P->D,P->pz);
    break;
  default:  error_message("Bad type to output point!!");
  }

  /* then output it */

  if (O->message != NULL) {
    switch (O->OT) {
    case TERM_M:
    case FILE_M:  fprintf(O->fp,"%s %g\n",O->message->str,pt);  /* don't worry about binary yet */
                break;
    case GUI_M:   printf("GUI output in the future\n");
    case NETCDF_M:   printf("NETCDF output in the future\n");
    }
  } else {
    switch (O->OT) {
    case TERM_M:
    case FILE_M:  fprintf(O->fp,"%g\n",pt);  /* don't worry about binary yet */
                  break;
    case GUI_M:   printf("GUI output in the future\n");
    case NETCDF_M:   printf("NETCDF output in the future\n");
    }
  }

#ifdef DEBUG2
  g_print("Finished with output_point()\n");
#endif /* DEBUG2 */

}


void output_depth_line(depth_line *P)
{
  output_options_t *O = P->OP;
  field1D *D;

  D = P->D;
  switch(O->OT) {
  case TERM_M: write_field1D_ascii(stdout,D);
    return;
  case FILE_M: 
    if (global_time >= O->start_time) {
      switch(O->FOT) {
      case ASCII_M: write_field1D_ascii(O->fp,D);
        return;
      case BINARY_M: write_field1D_binary(O->fp,D);
        return;
      default: error_message("BAD file type to output_depth_line()");
      }
    }
    break;
  default: error_message("BAD output type to output_depth_line()");
  }

}


void output_radius_line(radius_line *P)
{
  output_options_t *O = P->OP;
  field1D *D;

  D = P->D;
  switch(O->OT) {
  case TERM_M: write_field1D_ascii(stdout,D);
    return;
  case FILE_M: 
    if (global_time > O->start_time) {
      switch(O->FOT) {
      case ASCII_M: write_field1D_ascii(O->fp,D);
        return;
      case BINARY_M: write_field1D_binary(O->fp,D);
        return;
      default: error_message("BAD file type to output_radius_line()");
      }
    }
    break;
  default: error_message("BAD output type to output_radius_line()");
  }

}



void traverse_point_list(GSList *plist)
{
  GSList *list;

  for (list=plist;list;list=list->next)
    output_point((point * ) list->data);

}

void traverse_depth_list(GSList *clist)
{
  GSList *list;

  for (list=clist;list;list=list->next)
    output_depth_line((depth_line * ) list->data);

}

void traverse_radius_list(GSList *rlist)
{
  GSList *list;

  for (list=rlist;list;list=list->next)
    output_radius_line((radius_line * ) list->data);

}


void level1_output(oflags1 *l1)
{
  traverse_point_list(l1->point_list);
  traverse_depth_list(l1->depth_list);
  traverse_radius_list(l1->radius_list);

}
