/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* tke.c --- routines for the TKE dynamics */

#include <math.h>

#include "defs.h"
#include "tke.h"


/* Global Variables defined in defs.h */

/* Turbulence model constants */

const double Sq = 0.2;   // units ????? also next 3??
const double Sm = 0.39;   /* Momentum mixing coef. in MY model */
double l;
const double Bt = 16.6;   /* dissipation const. for MY models */

/* Global Field Variables */

field1D *QQ;        /* turb. velocity so that tke = q^2/2 */
field1D *BB;        /* tke b = q^2/2     */
field1D *LL;        /* turb. length scale   */
field1D *Q2L;       /* Mellor Yamada type length scale closure variable */

field1D *BBT1;
field1D *BBT2;

field1D *DISS;     // TKE dissipation = 2*q*b/(B*l)    (size N)
field1D *PROD;     // TKE shear production      (size N)
field1D *TKEDIFF; // TKE diffusion     (size N)
field1D *TKELOSS;  // TKE loss due to bubble breakup
field1D *KQV;       // vertical eddy viscosity = Sq*l*q (size N-1)

field1D *DDT1;     /* These are the d/dt of dissipation */
field1D *DDT2;

field1D *Q2LT1;     /* These are the d/dt of q^2 l MY equation */
field1D *Q2LT2;

field1D *ILD;      /* Used in the MY functions, just is 1/LL to save a division */
field1D *LLZ;      /* This is the fixed length scale the q^2l eq. needs to relax to */

/* Functions */


void file_load_tke_ic(field1D *B, field1D *Q, char *tkefile)
{

  int i, rval;

  rval = load_field1D_ascii(tkefile,B);
  if (rval == 0)
    error_message("tke file reading error");

  for (i=0;i<B->N;i++) 
    DR1(Q,i) = sqrt( 2.0 * DR1(B,i) );


}


void set_inverse_length_scale_for_diss(double Bt, field1D *LL)
{
  int i;

  for (i=0;i<nz;i++) {
    DR1(ILD,i) = 1.0/(Bt*DR1(LL,i));
  }

}


void  deallocate_tke_fields()
{

  deallocate_field1D(QQ);
  deallocate_field1D(BB);
  deallocate_field1D(BBT1);
  deallocate_field1D(BBT2);
  deallocate_field1D(LL);
  deallocate_field1D(KQV);
  deallocate_field1D(DISS);
  deallocate_field1D(TKEDIFF);
  deallocate_field1D(PROD);
  deallocate_field1D(TKELOSS);
  deallocate_field1D(DDT1);
  deallocate_field1D(DDT2);
  deallocate_field1D(Q2LT1);
  deallocate_field1D(Q2LT2);

  deallocate_field1D(ILD);
  deallocate_field1D(LLZ);

}


void  init_tke_fields(int nz, double z0b, double z0s)
{

   QQ = allocate_field1D(nz);
   BB = allocate_field1D(nz);
   LL = allocate_field1D(nz);
   Q2L = allocate_field1D(nz);
   DISS = allocate_field1D(nz);
   PROD = allocate_field1D(nz);
   TKEDIFF = allocate_field1D(nz);
   TKELOSS = allocate_field1D(nz);
   KQV = allocate_field1D(nz-1);

   BBT1 = allocate_field1D(nz-2);
   BBT2 = allocate_field1D(nz-2);
   DDT1 = allocate_field1D(nz-2);
   DDT2 = allocate_field1D(nz-2);
   Q2LT1 = allocate_field1D(nz-2);
   Q2LT2 = allocate_field1D(nz-2);

   ILD = allocate_field1D(nz);
   LLZ = allocate_field1D(nz);

   set_field1D_to_value(QQ,1e-3);
   set_field1D_to_value(BB,0.5e-6);

   set_field1D_to_value(DISS,1e-8);
   set_field1D_to_value(PROD,1e-8);


   /* Does the Craig and Banner length scale stuff */

#ifdef CB_LENGTH_SCALE
   set_CB_length_scale(LL, z0b, z0s);   // these are the bottom & surf roughness values
   set_CB_length_scale(LLZ, z0b, z0s);   // these are the bottom & surf roughness values
   //   set_parabola_length_scale(LL, z0b, z0s);   // these are the bottom & surf roughness values
   //   set_parabola_length_scale(LLZ, z0b, z0s);   // these are the bottom & surf roughness values
#endif
#ifdef WALL_LENGTH_SCALE
   set_wall_length_scale(LL, z0b);   // these are the bottom & surf roughness values
   set_wall_length_scale(LLZ, z0b);   // these are the bottom & surf roughness values
#endif

#ifdef SHAW_IC
   file_load_tke_ic(BB,QQ,"mTKEg.dat");
#endif /* SHAW_IC */
   /* set the q^2l variable to an appropriate value */

   copy_field1D_mult(Q2L,BB,2.0);     // q^2l = 2.0 * (q^2/2) * l 
   mult_field1D(Q2L,LL);

   
   set_field1D_to_value(BBT1,0.0);
   set_field1D_to_value(BBT2,0.0);

   set_inverse_length_scale_for_diss(Bt,LL);


}




/* ---------------------------------------------------------------------------------
   --------------- tke equations w/ fixed length scale -----------------------------
   ---------------------------------------------------------------------------------   */


/* Here the rhs of the tke equation except the diffusion which is don
   implicitly is calculated here, so we need dissipation, bflux, etc. */

void rhs_implicit_tke_eq(field1D *BT, const field1D *B, const field1D *Q, const field1D *DISS, 
                         const field1D *Kv, double *bbed)
{

  int i, N = B->N;

  double *bd = B->data+1;

  double *dissd = DISS->data+1;
  double *btd = BT->data;
  double *kv = Kv->data+1;

  double *prodd = PROD->data+1;
  double *diffd = TKEDIFF->data+1;
  double bb, bp;
  double diffusion, vgrad, vgrad1, vgrad2, shear_production = 0.0;
  double buoyancy_flux = 0.0;
  double dbdt;

  double kvp, kvm;
  double kvtop;
  double prod, qq;
  double ll02 = SQR(DR1(LL,0));
  double ll = DR1(LL,0);
  static double Smtrue = 0.39;


  /* first do the diffusion at the first grid cell */

  kvm = *kv++;

  for (i=1;i<N-1;i++) {

    kvp = *kv++;
    bb = *bd++;
    bp = *bd;

    dbdt = - *dissd++;

    /* next do the shear production term stuff */
    /*  \ovz (KUV [\delta_z(U)^2 +\delta_z(V)^2] ) */
    /* to be numerically consistent w/ the momentum diffusion term */
    /* But had to switch back to the original version because that gave right answers!! */

    /* VERSION 1 NO GOOD!!! */
    /*
    vgrad1 = DR1(KUV,i)*(SQR(DR1(UU,i+1)-DR1(UU,i)) + SQR(DR1(VV,i+1)-DR1(VV,i)));    
    vgrad2 = DR1(KUV,i-1)*(SQR(DR1(UU,i)-DR1(UU,i-1)) + SQR(DR1(VV,i)-DR1(VV,i-1)));    
    shear_production = idz2*0.125*(vgrad1+vgrad2);
    */
    /* VERSION 2,  KUv * \bar{\delta_z(U)^2} - this should be good but it's NOT! */
    /*
    vgrad1 = (SQR(DR1(UU,i+1)-DR1(UU,i)) + SQR(DR1(VV,i+1)-DR1(VV,i)));    
    vgrad2 = (SQR(DR1(UU,i)-DR1(UU,i-1)) + SQR(DR1(VV,i)-DR1(VV,i-1)));    
    shear_production = 0.25*idz2*Sm*DR1(LL,i)*DR1(QQ,i)*(vgrad1+vgrad2);
    */

    vgrad = 0.25*idz2*(SQR(DR1(UU,i+1)-DR1(UU,i-1)) + SQR(DR1(VV,i+1)-DR1(VV,i-1)));
    //    shear_production = (Sm * DR1(LL,i)*DR1(QQ,i)+1.2e-6) * vgrad;
    shear_production = 0.5*(kvm+kvp) * vgrad;

    *prodd++ = shear_production;
    dbdt += shear_production;

    /* next do buoyancy flux */

#ifdef BUOYANCY_FLUX
    buoyancy_flux = DR1(BFLUX,i);
    dbdt += buoyancy_flux;
#endif /* BUOYANCY_FLUX */

#ifdef SEDIMENT_BUOYANCY_FLUX
    //    dbdt += DR1(SBFLUX,i);
#endif /* SEDIMENT__BUOYANCY_FLUX */

    *btd++ = dbdt;
  }

  /* This is for the plunging wave tke source */

  if ((wave_breaking ==ON) && (turb_source == PLUNGING)) {
    btd = BT->data;
    for (i=1;i<N-1;i++) 
      *btd++ += DR1(TKE_SOURCE,i);
  }



#ifdef NO_FLUX_TKE_BOTTOM_BC
  DR1(BB,0) = DR1(BB,1);
#endif

#ifdef NO_SLIP_TKE_BOTTOM_BC
  DR1(BB,0) = 0.0;
#endif

#ifdef MY_TKE_BOTTOM_BC    /* shear prod = dissipation */
  vgrad = idz2*(SQR(DR1(UU,1)-DR1(UU,0)) + SQR(DR1(VV,1)-DR1(VV,0)));
  bb = 0.5*Smtrue * ll02 * Bt * vgrad;
  qq = sqrt(2.0*bb);
  prod = Smtrue * ll * qq * vgrad;
  DR1(PROD,0) = prod;
  DR1(PROD,N-1) = DR1(PROD,N-2);    // this is the surface condition!!  ACK
  *bbed = bb;
#endif


}




/* This function is run AFTER time step to convert between B and Q */

void calculate_tke_stats(field1D *Q, field1D *DISS, field1D *Kv, field1D *KUv, const field1D *B, const field1D *LL)
{

  int i;
  int N = B->N;
  double tmp;
  double *bd = B->data;
  double *qd = Q->data;
  double *dissd = DISS->data;
  double *kvd = Kv->data;
  double *kuvd = KUv->data;
  double *ild = ILD->data;

  for (i=0;i<N;i++) {
    *qd = sqrt((2.0*(*bd)));
    //*dissd++ = 2.0*(*qd++) * (*bd++)/(Bt * DR1(LL,i));
      *dissd++ = 2.0*(*qd++) * (*bd++) * (*ild++);
  }

  for (i=0;i<N-1;i++) {
    tmp = 0.25*(DR1(Q,i)+DR1(Q,i+1))*(DR1(LL,i)+DR1(LL,i+1));
    *kvd++ = Sq*tmp + 1.2e-6;
    *kuvd++ = Sm*tmp + 1.2e-6;
    }

}

/* This function tries to take into account the Sm dependence on the prod/diss ratio */

void calculate_tke_stats_new(field1D *Q, field1D *DISS, field1D *Kv, field1D *KUv, const field1D *B, const field1D *LL)
{

  int i;
  int N = B->N;
  double tmp;
  double *bd = B->data;
  double *qd = Q->data;
  double *dissd = DISS->data;
  double *kvd = Kv->data;
  double *kuvd = KUv->data;
  double *ild = ILD->data;
  double Smtmp;

  for (i=0;i<N;i++) {
    *qd = sqrt((2.0*(*bd)));
    //*dissd++ = 2.0*(*qd++) * (*bd++)/(Bt * DR1(LL,i));
    *dissd++ = 2.0*(*qd++) * (*bd++) * (*ild++);
  }

  for (i=0;i<N-1;i++) {   
    tmp = 0.25*(DR1(Q,i)+DR1(Q,i+1))*(DR1(LL,i)+DR1(LL,i+1));
    Smtmp = 0.6992 - 0.3059 *(DR1(PROD,i+1)+DR1(PROD,i))/(DR1(DISS,i)+DR1(DISS,i+1));
    if (Smtmp<0.1)
      Smtmp = 0.1;
    *kvd++ = Sq*tmp + 1.2e-6;
    *kuvd++ = Smtmp*tmp + 1.2e-6;
    }

  //  printf("** Sm = %g at the top i=n-2\n",Smtmp);
  //  printf("** PROD(n-1)=%g  PROD(n-2)=%g  DISS(n-1)=%g, DISS(n-2)=%g\n",DR1(PROD,N-1),DR1(PROD,N-2),DR1(DISS,N-1),DR1(DISS,N-2));
}



void set_const_Kv(field1D *Kv, double kv)
{

  int i, N = Kv->N;
  double *kvd = Kv->data;

  for (i=0;i<N-1;i++) 
    *kvd++ = kv;

}



void set_CB_length_scale(field1D *LL, double z0b, double z0top)
{
  const double kappa = 0.4;
  int N = LL->N;
  int i;
  double z, zprime, zmid = 0.5*(-z0b + depth +z0top);
  double l;

  for (i=0;i<N;i++) {
    z = i*dz + z0b;
    zprime = depth - i*dz + z0top;
    if (z<zprime) 
      DR1(LL,i) = kappa*z;
    else
      DR1(LL,i) = kappa*zprime;

  }

}


void set_parabola_length_scale(field1D *LL, double z0b, double z0top)
{
  const double kappa = 0.4;
  int N = LL->N;
  int i;
  double z, zprime;
  double l;

  for (i=0;i<N;i++) {
    z = i*dz + z0b;
    zprime = depth - i*dz + z0top;
    DR1(LL,i) = kappa* (z*zprime)/(z + zprime);
  }

}


void set_wall_length_scale(field1D *LL, double z0b)
{
  const double kappa = 0.4;
  int N = LL->N;
  int i;
  double z;


  for (i=0;i<N;i++) {
    z = i*dz;
    DR1(LL,i) = kappa*(z0b+z);;
  }
 

}



/* ---------------------------------------------------------------------------------
   --------------- k-epsilon equation functions ------ -----------------------------
   ---------------------------------------------------------------------------------   */




void rhs_implicit_kepsilon_eq(field1D *BT, field1D *DT, const field1D *B, const field1D *DISS, const field1D *PROD, 
                              const field1D *Kv, double *kbed, double *dissbed)
{

  int i, N = DISS->N;

  static const double Ce1 = 1.44;      // C_\epsilon1 in GOTM
  static const double Ce2 = 1.92;      // C_\epsilon2 in GOTM
  static const double Ce3 = -0.4;      //  this is if buoyancy effects are also included 
  static const double Sige =   1.077498753010693;
  static const double iSige =   9.280753200000001e-01;   //   this is 1/\sigma_\epsilon in GOTM

  static const double Cmu0 = 0.5562;
  static const double cube_Cmu0 =   1.720651643280000e-01;
  static const double i_quad_Cmu0 =     1.044903206465639e+01;  
  static const double i_sqr_Cmu0 =    3.232496259032079e+00;
  static const double iCmu02 = 3.232496259032079;
  static const double kappa = 0.4;
  static const double itwo32 = 0.35355339059327;  // 2^(-3/2)

  double ll02 = SQR(DR1(LL,0));
  double ll = DR1(LL,0);  //0.5*(DR1(LL,0)+DR1(LL,1));
  double tmp;

  double kvbed = DR1(Kv,0);
  double kvtop = DR1(Kv,N-2);

  double prod,  qq;
  double ustar2;

  double *bd = B->data+1;

  double *dissd = DISS->data+1;
  double *dtd = DT->data;
  double *btd = BT->data;
  double *kvd = Kv->data;

  double *prodd = PROD->data+1;

  double bb, bp, dd, dp;
  double vgrad, shear_production;
  double vgrad1, vgrad2;

  double Stke;

  double bubble_buoyancy_flux;
  double sediment_buoyancy_flux;
  double density_buoyancy_flux;
  
  double kv,kvm; 
  double ldiff, lprod, ldiss, lbflux, fac;

  double dddt, dbdt;

  //  double prod2, bb2;  /* for alternate bottom bc's that don't work */

  /* first do the diffusion at the first grid cell */

  kvm = *kvd++;
  dd = *dissd++;
  bb = *bd++;

  for (i=1;i<N-1;i++) {

  /* The tke diffusion */
  /* added in the variable length scale stuff */

    bp = *bd++;
    dp = *dissd++;

    kv = *kvd++;

    dbdt = -dd;


    vgrad1 = kv * (SQR(DR1(UU,i+1)-DR1(UU,i)) + SQR(DR1(VV,i+1)-DR1(VV,i)));    
    vgrad2 = kvm * (SQR(DR1(UU,i)-DR1(UU,i-1)) + SQR(DR1(VV,i)-DR1(VV,i-1)));    
    shear_production = idz2*0.5*(vgrad1+vgrad2);

    /*
    vgrad = 0.25*idz2*(SQR(DR1(UU,i+1)-DR1(UU,i-1)) + SQR(DR1(VV,i+1)-DR1(VV,i-1)));
    shear_production = 0.5*(kvm+kv) * vgrad;   // This is NOW fixed 
    */
    kvm = kv;

    *prodd++ = shear_production;
    dbdt += shear_production;

    /* next do buoyancy flux */

#ifdef BUBBLE_BUOYANCY_FLUX
    buoyancy_flux = DR1(BFLUX,i);
    dbdt += bubble_buoyancy_flux;
#endif /* BUBBLE_BUOYANCY_FLUX */

#ifdef SEDIMENT_BUOYANCY_FLUX
    sediment_buoyancy_flux = DR1(SBFLUX,i);
    dbdt += sediment_buoyancy_flux;
#endif /* SEDIMENT_BUOYANCY_FLUX */

#ifdef DENSITY_BUOYANCY_FLUX
    density_buoyancy_flux = DR1(WRHOFLUX,i);
    dbdt += density_buoyancy_flux;
#endif /* DENSITY_BUOYANCY_FLUX */

    *btd++ = dbdt;

    /* --------------------------------------- */
    /* --- Next do the epsilon equation -----  */
    /* --------------------------------------- */

  /* The dissipation diffusion - also done slowly */
  /* added in the variable length scale stuff */

    fac = dd/bb;

    ldiss = Ce2 * dd;

    lprod = Ce1 * shear_production;

#ifdef DENSITY_BUOYANCY_FLUX
    lbflux = Ce3 * density_buoyancy_flux;
    lprod += lbflux;
#endif /* DENSITY_BUOYANCY_FLUX */

    dddt = (lprod - ldiss)*fac;

    *dtd++ = dddt;

    dd = dp;
    bb = bp;
  }

  /* This is for the plunging wave tke source */

  if ((wave_breaking ==ON) && (turb_source == PLUNGING)) {
    btd = BT->data;
    dtd = DT->data;
    dissd = DISS->data+1;
    bd = B->data+1;
    for (i=1;i<N-1;i++) {
      Stke = DR1(TKE_SOURCE,i);
      *btd++ += Stke;
      *dtd++ += (*dissd++) * Stke * Ce1 / (*bd++);
    }
  }


  /* For the turb it's a PROD = DISS balance at bed -- assumes a log bl*/
  /* This B-C comes from GOTM manual too */

  vgrad = idz*sqrt(SQR(DR1(UU,1)-DR1(UU,0)) + SQR(DR1(VV,1)-DR1(VV,0)));
  ustar2 = kvbed*vgrad;


  bb = ustar2 * i_sqr_Cmu0;  // ustart2/(Cmu0*Cmu0); OLD BURCHARD BC CRAP - supposed log V structure!
  //  bb2 = i_sqr_Cmu0 * SQR(ll) * SQR(vgrad);   // PROD == DISS BC
  if (bb<1e-8)                 // minimum value of bed turbulence!
    bb=1e-8;

  qq = sqrt(bb);
  //prod2 = Cmu0 * qq * ll * SQR(vgrad);
  prod = ustar2 * vgrad;

  /* ---> This was the old MY type BC
  //  bb = ll02 * vgrad * iCmu02;
  //  printf("** bb = %g  qq = %g \n",bb,qq);
  //  prod = Cmu0 * qq * ll * vgrad;
  */

  DR1(PROD,0) = prod;


  *kbed = bb;

  DR1(PROD,N-1) = DR1(PROD,N-2);    // this is the surface condition!!  ACK

  /* Dirichelet boundary conditions */
  /* This is for the bottom epsilon boundary condition */

  //  *dissbed = prod;    // old crappy bc cube_Cmu0 * CUBE(qq)/ll;
  *dissbed = cube_Cmu0 * CUBE(qq)/ll;

  //printf("** prod=%g, prod2=%g, bb=%g, bb2=%g, diss_bed=%g\n",prod,prod2,bb,bb2,*dissbed);

}




void calculate_kepsilon_stats(field1D *BB, field1D *DISS, field1D *Kv, field1D *KUv)
{

  int i;
  int N = BB->N;
  double tmp;
  double *bd = BB->data;
  double *dissd = DISS->data;
  double *kvd = Kv->data;
  double *kuvd = KUv->data;

  static double Cmu4 =  9.570264439923362e-02;  //0.09490138671180;  --> Cmu^4 

  double bb, dd;
  double kup, kdn, kavg;

  bb = *bd++;
  dd = *dissd++;

  kdn = Cmu4 * bb*bb/dd;

  for (i=0;i<N-1;i++) {
    bb = *bd++;
    dd = *dissd++;
    /*    if (dd==0.0) 
	  printf("** Warning dd==0.0 in calc_kepsilon_stats"); */
    kup = Cmu4 * bb * bb/dd;
    kavg = 0.5*(kup+kdn);

#ifdef DEBUG
    if (kavg>10){
      printf("** WARNING kavg>10:  kavg=%g  at i=%d, kup=%g, kdn=%g, eps=%g, tke=%g\n",kavg,i,kup,kdn,dd,bb);
      kavg = 10.0;
    }
#endif /* DEBUG */

    kdn = kup;
    *kuvd++ = kavg;      // no 1.2e-6 for background
    *kvd++ = kavg;       // 
    }

}


/* check dissipation limits in stratification */

void check_dissipation_stratified(field1D *DISS, field1D *N2)
{
  int N = DISS->N;
  int i;

  double *dissd=DISS->data;
  double *N2d = N2->data;
  double *tked = BB->data;

  double N2surf = 0;
  double N2bed = 0;
  double diss2min, diss, diss2;
  double tke, tke2;
  double N2p, N2m, N2avg;

  // surface and near bed N2 = 0

  N2m = N2bed;

  for (i=0;i<N-1;i++) {

    N2p = *N2d++;
    N2avg = 0.5* (N2p+N2m);

    diss = *dissd;
    diss2 = diss*diss;
    tke = *tked++;
    tke2 = tke*tke;
    diss2min = 0.045*tke2*N2avg;

    if (diss2 < diss2min) 
      *dissd = sqrt(diss2min);
         
    dissd++;
    N2m = N2p;

  }

  /* Near surface N - this could be non-zero if there is a buoyancy source */

  N2p = N2surf;

  N2avg = 0.5* (N2p+N2m);

  diss = *dissd;
  diss2 = diss*diss;
  tke = *tked++;
  tke2 = tke*tke;
  diss2min = 0.045*tke2*N2avg;

  if (diss2 < diss2min)
    *dissd = sqrt(diss2min);
         

}





