/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* 
   wavemodel.c 

   functions for calculating wave paramters:
   wavenumber, cg, wave dissipation

*/


#include <stdio.h>
#include <math.h>

#include "defs.h"
#include "wavemodel.h"

/* Wave math functions */

double tanh(double x)
{
  double t;

  t = sinh(x)/cosh(x);
  return t;

}

double sech(double x)
{
  double t;

  t = 1.0/cosh(x);
  return t;

}


double get_wavenum(double omega, double depth)
{

  double f, dfdk;
  double k, kh;


  k = omega/(grav*depth);
  kh = k*depth;
  f = grav*k*tanh(kh) - SQR(omega);

  while (fabs(f)>1e-10) {
    kh = k*depth;
    dfdk = grav*kh*(SQR(sech(kh)) + grav*tanh(kh));
    k = k - f/dfdk;
    f = grav*k*tanh(kh) - SQR(omega);
  }

  return k;

}

double get_cg(double k, double depth)
{

  double denom;
  double numerator;
  double cg;
  double kh = depth*k;

  denom = sqrt(grav*k*tanh(kh));
  numerator = 0.5*(grav*tanh(kh)+grav*kh*SQR(sech(kh)));

  cg = numerator/denom;

  return cg;

}

/* 
   Wave dissipation function given in Church & Thornton, without rho 
   d/dx(E cg) = epsilon.    The units of E are h*u^2 = m^3/s^2, cg -> m/s
   so epsilon is m^3/s^3,  or   W*m/kg.   If we multiplied by rho -> kg/m^3,
   then we would have W/m^2.    But we don't include rho.   This is also
   the average dissipation over a wave period.    So really it's
   \bar{epsilon} = (1/T) \int epsilon dt.    The amount of energy put into
   the system over 1 wave cycle is T*epsilon.   
*/


double wave_averaged_energy_dissipation(double Hrms, double h, double freq, double gam, double B)
{

  double M1, M2, C1, tmp1, tmp2;
  double epsilon;

  tmp1 = 8.0 * (Hrms/(gam*h) - 1.0);
  M1 = 1.0 + tanh(tmp1);

  tmp1 = 1.0 + SQR(Hrms/(gam*h));
  tmp2 = pow(tmp1,2.5);
  M2 = 1.0 - 1.0/tmp2;

  C1 =(3.0/16.0)*sqrt(pi)*grav*freq*B*B*B;

  epsilon = C1 * CUBE(Hrms) * M1 * M2 / h;

#ifdef DEBUG
  printf("** wavemodel.c: averaged wave dissipation = %g\n",epsilon); 
#endif

  return epsilon;

}
